#!/usr/bin/python
from __future__ import print_function

from sys import stdout
from os import path, getcwd, listdir, chdir, makedirs, remove
from subprocess import call
from io import FileIO
from glob import glob
from functools import wraps
from contextlib import contextmanager
from runana import input_file_handling
from operator import add, mul
from runana.read_numbers import ignored

try:
    from operator import div
except ImportError:
    from operator import truediv as div

try:
    basestring          # Python 2.x
except NameError:
    basestring = str    # Python 3.x


OPERATIONS = {'add': add, 'mul': mul, 'div': div}


@contextmanager
def cwd(path):
    """ Contextmanager that changes working directory temporarily """
    oldpwd = getcwd()
    chdir(path)
    try:
        yield
    finally:
        chdir(oldpwd)


def generate_seq(start, incr, nvalues=0, incr_func=add):
    """Iterator that returns a sequence of numbers

    :param incr_func: function used to increment the return value. Can be one
        of the strings 'add', 'mul' or 'div'
    :type incr_func: func or str
    """
    if isinstance(incr_func, basestring):
        incr_func = OPERATIONS[incr_func]
    value = start
    yield value
    for i in range(1, nvalues):
        value = incr_func(value, incr)
        yield value


def generate_list(*args, **kwargs):
    """Wrap of generate_seq that returns a list instead of an iterator"""
    return list(generate_seq(*args, **kwargs))


class Dirs(object):
    """Container class for names of directories

    :param str scratch_base: Directory prefix
    :param str local_scratch_base: Prefix for directory in which programs
        are run. If `None` then `scratch_base` is used
    :param list copy_2_scratch: List of strings that are globbed and copied
        from the local scratch directory to scratch directory
    """
    def __init__(self, scratch_base, local_scratch_base=None,
                 copy_2_scratch=['*.txt', '*.nml', '*.stdout', '*.dat']):
        if local_scratch_base is None:
            local_scratch_base = scratch_base
        self.scratch_base = scratch_base
        self.local_scratch_base = local_scratch_base
        self.copy_2_scratch = copy_2_scratch
        makedir(self.scratch_base)


def makedir(dir_):
    if not path.exists(dir_):
        makedirs(dir_)


class OpenWithNone(FileIO):
    def __init__(self, file_string, *args, **kwargs):
        self.file_string = file_string
        if file_string:
            super(OpenWithNone, self).__init__(file_string, *args, **kwargs)

    def __enter__(self):
        handle = None
        if self.file_string:
            handle = super(OpenWithNone, self).__enter__()
        self.fd = handle
        return self.fd

    def __exit__(self, type, value, traceback):
        if self.fd:
            super(OpenWithNone, self).__exit__()


def replace_string_in_file(fileName, text_to_search, text_to_replace):
    with open(fileName, 'r') as file_handle:
        filedata = file_handle.read()
    filedata = filedata.replace(text_to_search, text_to_replace)
    with open(fileName, 'w') as file_handle:
        file_handle.write(filedata)


def run_program(program, cmdargs, stdin_f, stdout_f, stderr_f,
                run=True, cmd_prepend="", run_from_cmd=True,
                **kwargs):
    """Runs `program` with `cmdargs` using `subprocess.call`.
    :param str stdin_f: File from which to take standard input
    :param str stdout_f: File in which to put standard output
    :param str stderr_f: File in which to put standard error
    :param bool run: Whether to actually run `program`
    If `True` the program return code is returned.
    If false a string pointing to the script which will run
 the program is returned
    :param str cmd_prepend: Put in the beginning of the bash script
    :param bool run_from_cmd: Run `program` using the generated bash
    script instead of running it directly

    """
    time_file_name = '.'.join(stdout_f.split('.')[:-1])+'.time'
    cmd_file_name = '.'.join(stdout_f.split('.')[:-1])+'.sh'
    with open(cmd_file_name, 'w') as cmd_file:
        cmd = ' '.join([program]+cmdargs)
        time_cmd = "/usr/bin/time -o {time_file}".format(time_file=time_file_name)
        cmd = "{time_cmd} {cmd} 1> {stdout} 2> {stderr} \n".format(time_cmd=time_cmd,
                                                                   cmd=cmd,
                                                                   stdout=stdout_f,
                                                                   stderr=stderr_f)
        cmd = cmd_prepend + cmd
        cmd_file.write(cmd)
    if run:
        with OpenWithNone(stdin_f, 'r') as input_file,  open(stdout_f, 'w') as stdout_file, open(stderr_f, 'w') as stderr_file:
            if run_from_cmd:
                retcode = call(["bash", cmd_file_name], **kwargs)
            else:
                try:
                    with open(time_file_name, 'w') as time_file:
                        with print_time(time_file):
                            retcode = call([program]+cmdargs, stdin=input_file,
                                           stdout=stdout_file, stderr=stderr_file, **kwargs)
                except Exception as e:
                    print(e)
                    print('program ', program)
                    print('cmdargs', cmdargs)
                    print('stdin   ', stdin_f)
                    print('stdout  ', stdout_f)
                    print('stderr  ', stderr_f)
                    # print 'kwargs  ', kwargs
                    print(getcwd())
                    raise
                
        replace_string_in_file(stdout_f, '\r', '\n')
        return retcode
    else:
        return cmd_file_name


from subprocess import Popen, PIPE
from time import sleep
# from time import time
# from os import fsync
def run_program_print_output(program, cmdargs,
                             stdin_f, stdout_f, stderr_f, print_output=False,
                             **kwargs):
    with OpenWithNone(stdin_f, 'r') as input_file:
        with open(stdout_f, 'w', buffering=0) as stdout_file:
            with open(stderr_f, 'w', buffering=0) as stderr_file:
                try:
                    with open(stdout_f+'.sh', 'w') as cmd_file:
                        cmd_file.write(' '.join([program]+cmdargs))
                    if print_output:
                        # start_time = time()
                        process = Popen([program]+cmdargs, stdin=input_file,
                                        stdout=PIPE, stderr=PIPE, bufsize=0, **kwargs)
                        # print(time()-start_time)
                        # print('Right after process call')
                        while True:
                            stdout = process.stdout.readline()
                            stderr = process.stderr.readline()
                            if stdout == '' and stderr == '' and process.poll() is not None:
                                break
                            if stdout:
                                # print(time()-start_time)
                                print(stdout, end='')
                                stdout_file.write(stdout)
                                # stdout_file.flush()
                                # fsync(stdout_file.fileno())
                            if stderr:
                                print(stderr, end='')
                                stderr_file.write(stderr)
                            sleep(0.1)
                    else:
                        call([program]+cmdargs, stdin=input_file,
                             stdout=stdout_file, stderr=stderr_file, bufsize=0, **kwargs)
                except Exception as e:
                    print(e)
                    print('program ', program)
                    print('cmdargs', cmdargs)
                    print('stdin   ', stdin_f)
                    print('stdout  ', stdout_f)
                    print('stderr  ', stderr_f)
                    # print 'kwargs  ', kwargs
                    print(getcwd())
                    raise
    replace_string_in_file(stdout_f, '\r', '\n')


def name_stdout(program, add=''):
    if isinstance(program, basestring):
        prog = program
    else:
        prog = program[0]
    stdouts = prog.split('/')[-1]
    # stdouts = stdouts.split('.')[0].split('_')[0]+add+'.std'
    stdouts = stdouts+add+'.std'
    return stdouts+'out', stdouts+'err'


def run_prog(program, cmdargs=[], stdin_f=None, add='', **kwargs):
    # run_program_print_output(program, cmdargs, stdin_f,
    return run_program(program, cmdargs, stdin_f,
                       *name_stdout(program, add), **kwargs)


def copy_ignore_same(from_file, to_file):
    from shutil import copy, Error
    try:
        copy(from_file, to_file)
    except Error as err:
        with open('shutil.err', 'a') as f1:
            f1.write(str(err)+'\n')
    except IOError:
        pass


def copy_to_scratch(WorkDir, file_strings):
    files = []
    for file_string in file_strings:
        files = files+glob(file_string)
    for fil in files:
        copy_ignore_same(fil, WorkDir)


@ignored(OSError)
def get_subdirs(a_dir='./'):
    return [name for name in listdir(a_dir)
            if path.isdir(path.join(a_dir, name))]


def generate_run_ID(work_dir, invalid_IDs=[], prepend=""):
    subdirs = get_subdirs(work_dir)
    ID = 1
    while True:
        strID = str(ID)
        if prepend:
            strID = prepend + strID
        if strID in subdirs or strID in invalid_IDs:
            ID = ID+1
        else:
            return strID


def make_run_dirs(ScratchBase, LScratchBase, **gen_ID_kwargs):
    ID = generate_run_ID(ScratchBase, **gen_ID_kwargs)
    work_dir = path.join(ScratchBase, ID)
    lwork_dir = path.join(LScratchBase, ID)
    makedir(work_dir)
    makedir(lwork_dir)
    return ID, work_dir, lwork_dir


def lock_wrap_retry(Dir, nretries=10, wait=0.1):
    def decorate(f):
        @wraps(f)
        def call(*args, **kwargs):
            import fcntl, time
            pid_file = path.join(Dir, 'lock_ID_gen.pid')
            with open(pid_file, 'w') as lock_fp:
                for attempt in range(nretries):
                    try:
                        fcntl.lockf(lock_fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
                    except IOError:
                        import random
                        print('Waiting for lock on', Dir, attempt)
                        time.sleep(wait+0.05*random.random())
                        continue
                    break
                else:
                    print('Failed to get ID generation lock')
                try:
                    ret = f(*args, **kwargs)
                    fcntl.lockf(lock_fp, fcntl.LOCK_UN)
                finally:
                    fcntl.lockf(lock_fp, fcntl.LOCK_UN)
            return ret
        return call
    return decorate


def save_info_in_file(filename, command, copy_back=None):
    with open(filename, 'w') as output_file:
        call(command, stdout=output_file)
    if copy_back:
        copy_ignore_same(filename, copy_back)


def calc_all(replacements, dirs, inp_file, programs,
             print_finish=True, filter_func='f90nml', use_stdin=False,
             use_inp_file=True,
             **gen_ID_kwargs):

    base_dir = dirs.scratch_base
    dirID, work_dir, lwork_dir = lock_wrap_retry(base_dir, nretries=10, wait=0.1)(
        make_run_dirs)(base_dir, dirs.local_scratch_base, **gen_ID_kwargs)

    inp_file_local = path.join(work_dir, path.basename(inp_file))
    input_file_handling.INP_FILE_FILTERS[filter_func](inp_file, inp_file_local, replacements)

    with cwd(lwork_dir):
        save_info_in_file('hostname.txt', 'hostname', work_dir)
        save_info_in_file('started.txt', 'date', work_dir)
        inp_file_relative = path.relpath(inp_file_local, lwork_dir)
        run_core(programs, inp_file_relative, use_stdin=use_stdin,
                 use_inp_file=use_inp_file)
        copy_to_scratch(work_dir, dirs.copy_2_scratch)
        save_info_in_file('ended.txt', 'date', work_dir)
    if print_finish:
        print('Finished', dict((is_it_tuple(key),elem) for key,elem in replacements.items()))
    return dirID


def is_it_tuple(it):
    if isinstance(it, tuple):
        return it[1]
    else:
        return it


@contextmanager
def NamedTempFile(name, mode="w", **kwargs):
    """Contextmanager for creating a named temporary file """
    try:
        with open(name, mode=mode, **kwargs) as f:
            yield f
    finally:
        with ignored(FileNotFoundError):
            remove(name)


def run_core(programs, inp_file_relative, use_stdin=False,
             use_inp_file=True, add_temp_ignore_file=True):
    if add_temp_ignore_file:
        with NamedTempFile("ignore"):
            run_core_inner(programs, inp_file_relative, use_stdin,
                           use_inp_file)
    else:
        run_core_inner(programs, inp_file_relative, use_stdin,
                       use_inp_file)


def run_core_inner(programs, inp_file_relative, use_stdin=False,
                   use_inp_file=True):
    for program in programs:
        if hasattr(program, '__call__'):
            if use_inp_file:
                program(inp_file_relative)
            else:
                program()
        else:
            if use_inp_file:
                if use_stdin:
                    run_prog(program, [], stdin_f=inp_file_relative)
                else:
                    run_prog(program, [inp_file_relative])
            else:
                run_prog(program)


def add_to_fname(fname, add=''):
    fname_list = fname.split('.')
    fname_list[-2] = fname_list[-2]+add
    return '.'.join(fname_list)


def rerun(replacements, lworkdir, inp_file, programs, filter_func='f90nml'):
    inp_file_replace = add_to_fname(inp_file, '_rerun')
    with cwd(lworkdir):
        input_file_handling.INP_FILE_FILTERS[filter_func](inp_file,
                                                          inp_file_replace,
                                                          replacements)
        save_info_in_file('re_hostname.txt', 'hostname')
        save_info_in_file('restarted.txt', 'date')
        run_core(programs, inp_file_replace, lworkdir)
        save_info_in_file('re_ended.txt', 'date')


def merge_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


def product_replacements(product_iters):
    from itertools import product
    for value_set in product(*product_iters.values()):
        yield dict(zip(product_iters.keys(), value_set))


def co_replacements(co_iters):
    if len(co_iters) >= 1:
        for vector in zip(*co_iters.values()):
            yield dict(zip(co_iters.keys(), vector))
    else:
        yield {}


def chain_replacements(chain_iters):
    if len(chain_iters) >= 1:
        for key, chain_iter in chain_iters.items():
            for value in chain_iter:
                yield {key: value}
    else:
        yield {}


def replace_iter_gen(product_iters={}, chain_iters={}, co_iters={},
                     just_replace={}):
    for prod_iter in product_replacements(product_iters):
        for chain_iter in chain_replacements(chain_iters):
            for co_iter in co_replacements(co_iters):
                yield merge_dicts(merge_dicts(merge_dicts(just_replace,
                                                          prod_iter),
                                              chain_iter),
                                  co_iter)


def check_dirs(dirs):
    if not isinstance(dirs, Dirs):
        dirs = Dirs(dirs)
    return dirs

from functools import partial
def pick_filter_func(filter_func, calc_all):
    return partial(calc_all,filter_func=filter_func)


def execute(programs, input_file, dirs,
            chain_iters={}, product_iters={}, co_iters={}, just_replace={},
            filter_func='f90nml', use_stdin=False,
            calc_all=calc_all, **kwargs):
    """Run sequence of programs with different parameters defined by iters.

    :param list programs: List of strings with names of programs. Should
        contain absolute paths. Could alternately contain functions

    :param str input_file: Input file

    :param runana.run.Dirs dirs: Base directory in which programs will be run
    :type dirs: str or runana.run.Dirs

    :param dict chain_iters: Entries of the form
        {'Name of parameter':[*values to replace with*]}

    :param dict product_iters: Like `chain_iters`, but runs all combinations

    :param dict co_iters: Runs with several parameters changing simultanously

    :param dict just_replace: Entries of the form
        {'Name of parameter':*value to replace with*}

    :param str filter_func: Which filter function to use. Options are listed
        as keys in the INPUT_FILE_FILTERS dictionary

    :param bool use_stdin: send in the content of the filtered input file
        through stdin rather passing the name of the input file as the
        first command line argument

    :param func calc_all: Hook for the parallel decorator, please ignore
         this argument
    """
    dirs = check_dirs(dirs)
    input_file = path.abspath(input_file)
    calc_all = pick_filter_func(filter_func, calc_all)
    dir_IDs = []
    for replacers in replace_iter_gen(product_iters=product_iters,
                                      chain_iters=chain_iters,
                                      co_iters=co_iters,
                                      just_replace=just_replace):
        dir_ID = calc_all(replacers, dirs, input_file, programs,
                          use_stdin=use_stdin, **kwargs)
        dir_IDs.append(dir_ID)
    return dir_IDs


def execute_lock_par(lock, parallel, *args, **kwargs):
    """ Convenience function for running execute with a lock and/or in parallel """
    execute_here = execute
    if parallel:
        execute_here = parallel_wrap(parallel)(execute_here)
    if lock:
        execute_here = lock_wrap(lock)(execute_here)
    return execute_here(*args, **kwargs)


def common_start(chain_iters, just_replace):
    """ Returns modified `chain_iters` and `just_replace` such that the
 calculations will start at the first value of each variable in chain_iter
    """
    chain_iters_out = chain_iters.copy()
    replacers = {}
    if len(chain_iters_out) > 0:
        for (key, elem) in chain_iters_out.items():
            replacers[key] = elem.pop(0)
        else:
            elem.insert(0, replacers[key])
    just_replace = just_replace.copy()
    just_replace.update(replacers)
    return chain_iters_out, just_replace


class PoolApplyAsyncWrap(object):
    def __init__(self, pool):
        self.pool = pool

    def __call__(self, fun):
        def wrapped_f(*args, **kwargs):
            import copy
            for arg in args:
                print(arg)
                copy.deepcopy(arg)
            args = copy.deepcopy(args)
            kwargs = copy.deepcopy(kwargs)
            ret = self.pool.apply_async(fun, args, kwargs)
            return ret
        return wrapped_f


@contextmanager
def multi_stuff(parallel, kwargs):
    import multiprocessing
    import signal

    def initializer():
        """Ignore CTRL+C in the worker process."""
        signal.signal(signal.SIGINT, signal.SIG_IGN)

    number_of_cpus = multiprocessing.cpu_count()
    pool = multiprocessing.Pool(number_of_cpus, initializer=initializer)

    if parallel == 'Calc all':
        kwargs['calc_all'] = PoolApplyAsyncWrap(pool)(calc_all)
    elif parallel == 'auto_converge_var':
        if 'auto_converge_var' in kwargs:
            kwargs['auto_converge_var'] = PoolApplyAsyncWrap(pool)(kwargs['auto_converge_var'])
        else:
            kwargs['auto_converge_var'] = PoolApplyAsyncWrap(pool)(auto_converge_var)
    try:
        yield kwargs
        pool.close()
        pool.join()
    except KeyboardInterrupt:
        print("Caught KeyboardInterrupt, terminating workers")
        pool.terminate()
        pool.join()
        raise


import collections
def map_nested_dicts(dict_, func):
    if isinstance(dict_, collections.Mapping):
        ret = dict((k, map_nested_dicts(v, func)) for k, v in dict_.items()) 
    else:
        ret = func(dict_)
    return ret


def parallel_wrap(parallel=None):
    def decorate(fun):
        @wraps(fun)
        def call(*args, **kwargs):
            if parallel is None:
                converged_parameters = fun(*args, **kwargs)
            else:
                with multi_stuff(parallel, kwargs) as kwargs:
                    converged_parameters = fun(*args, **kwargs)
                    if converged_parameters:
                        converged_parameters = list(map(lambda x: x.get(), converged_parameters))
                        # converged_parameters = map_nested_dicts(converged_parameters, lambda x: x.get())
                # converged_parameters = multi_stuff(parallel, fun, args, kwargs)
            return converged_parameters
        return call
    return decorate


def lock_wrap(dir_):
    def decorate(fun):
        @wraps(fun)
        def call(*args, **kwargs):
            import fcntl
            makedir(dir_)
            pid_file = path.join(dir_, 'lock.pid')
            with open(pid_file, 'w') as lock_fp:
                try:
                    fcntl.lockf(lock_fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
                except IOError:
                    print('Another instance is running')
                    print('Lock file ', pid_file)
                    raise SystemExit
                try:
                    ret = fun(*args, **kwargs)
                    fcntl.lockf(lock_fp, fcntl.LOCK_UN)
                finally:
                    fcntl.lockf(lock_fp, fcntl.LOCK_UN)
            return ret
        return call
    return decorate


INTERVALS = (
    ('weeks', 604800),  # 60 * 60 * 24 * 7
    ('days', 86400),    # 60 * 60 * 24
    ('hours', 3600),    # 60 * 60
    ('minutes', 60),
    ('seconds', 1),
    )


def display_time(seconds, granularity=2, intervals=INTERVALS):
    list_ = display_time_list(seconds, granularity=granularity,
                              intervals=intervals)
    return ', '.join(list_)
    # result = []
    # for name, count in intervals:
    #     value = seconds // count
    #     if value:
    #         seconds -= value * count
    #         if value == 1:
    #             name = name.rstrip('s')
    #         result.append("{0:.0f} {1}".format(value, name))
    # return ', '.join(result[:granularity])


def display_time_list(seconds, granularity=2, intervals=INTERVALS):
    result = []
    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip('s')
            result.append("{0:.0f} {1}".format(value, name))
    return result[:granularity]


@contextmanager
def print_time(file_=stdout):
    """ Contextmanager that prints how much time was spent in it"""
    import time
    start = time.time()
    yield
    end = time.time()
    print(display_time(end-start), file=file_)


def rel_err_rel_var(O1, O2, x1, x2):
    """ Estimate of relative error `abs(x2/(x2-x1)*(O1-O2)/O2)` """
    return abs(x2/(x2-x1)*(O1-O2)/O2)

class ConvCrit(object):
    """ Contains information on how to check for convergence.

    :parameter func data_read: Function which will be executed in the directory in which the programs was run. It should return the observable in terms of which convergence is sought

    :parameter float eps: Desired precision

    :parameter func conv_funv: Function that calculates convergence criterion. It should take 4 arguments, `f(O1,O2,x1,x2)`, where `x1` and `x2` are the values of the numerical parameter at the current and previous calculation and `O1` and `O2` are the corresponding observable values

    """
    def __init__(self, data_read, eps = 1.0e-3, conv_func = rel_err_rel_var, itermax=10, common_start=False):
        self.eps = eps
        self.conv_func = conv_func
        self.data_read = data_read
        self.itermax = itermax
        self.common_start = common_start


def make_run_string(replacers):
    replacers = dict((is_it_tuple(key), elem) for key, elem in replacers.items())
    return str(replacers)


def float_also_array(varvalue):
    try:
        varvalue_f = float(varvalue)
    except TypeError:
        varvalue_f = varvalue[0]
    return varvalue_f


def auto_converge_var(var_name, var, replacements, dirs, inp_file, programs, conv_crit):
    data_read = conv_crit.data_read
    prevdirID = getattr(var, 'dirID', None)
    yield_diff = 1.0
    iteration = 1
    if prevdirID is None:
        iteration = 0
    for iteration, varvalue in enumerate(var, start=iteration):
        replacers = merge_dicts(replacements,{var_name:varvalue})
        dirID = calc_all(replacers, dirs, inp_file, programs, print_finish=False)
        if not prevdirID is None:
            # print(prevdirID,varvalue_prev,varvalue,data_read(path.join(dirs.local_scratch_base, dirID))
            with cwd(path.join(dirs.local_scratch_base, prevdirID)):
                O1 = data_read()
            with cwd(path.join(dirs.local_scratch_base, dirID)):
                O2 = data_read()
            yield_diff = conv_crit.conv_func(O1,O2,
                                             float_also_array(varvalue_prev),
                                             float_also_array(varvalue))
        try:
            print('{: <10.10}\t {:.2f}\t {:.2e}\t {}\t {}'.format(var_name[1], varvalue,
                                                                  yield_diff, iteration, dirID))
        except ValueError:
            print('{: <10.10}\t {}\t {:.2e}\t {}\t {}'.format(var_name[1], varvalue,
                                                                  yield_diff, iteration, dirID))
        if (abs(yield_diff) < conv_crit.eps):
            break
        prevdirID = dirID
        varvalue_prev = varvalue
    return {'VarValue': varvalue, 'PrevVarValue': varvalue_prev, 'iteration': iteration+1}


def auto_conv_sub(chain_iters,replacers, dirs, inp_file, programs, conv_crit, auto_converge_var):
    results={}
    for chain_iter in chain_iters:
        results[chain_iter] = auto_converge_var(chain_iter, chain_iters[chain_iter],
                                                replacers, dirs, inp_file, programs, conv_crit)
    return results


def auto_conv(programs, inp_file, dirs, conv_crit, chain_iters,
              product_iters={}, co_iters={}, just_replace={},
              auto_converge_var=auto_converge_var, auto_conv_sub=auto_conv_sub):
    """ Run programs until converged or chain_iters is exhausted.

    :param list programs: List of strings with names of programs. Should contain absolute paths. Could alternately contain functions

    :param str input_file: Input file

    :param runana.run.Dirs dirs: Base directory in which programs will be run
    :type dirs: str or runana.run.Dirs

    :param runana.run.ConvCrit conv_crit: Object specifying type of convergence

    :param dict chain_iters: Entries of the form {'Name of parameter':[*values to replace with*]}

    :param dict product_iters: Like `chain_iters`, but runs all combinations

    :param dict co_iters: Runs with several parameters changing simultanously

    :param bool use_stdin: send in the content of the filtered input file through stdin rather passing the name of the input file as the first command line argument
    """
    # :param dict just_replace: Entries of the form {'Name of parameter':*value to replace with*}
    # :param str filter_func: Which filter function to use. Options are listed as keys in the INPUT_FILE_FILTERS dictionary
    dirs = check_dirs(dirs)
    inp_file = path.abspath(inp_file)
    results = {}
    for replacers in replace_iter_gen(product_iters=product_iters,
                                      co_iters=co_iters,
                                      just_replace=just_replace):
        run_string = make_run_string(replacers)
        results[run_string] = auto_conv_sub(chain_iters,replacers, dirs, inp_file, programs,
                                        conv_crit, auto_converge_var)
    return results


def intercept_argument(args, kwargs, iarg, name):
    if len(args)>iarg:
        ret = args[iarg]
    else:
        ret = kwargs.get(name)
    return ret

def reinsert_argument(args, kwargs, iarg, name, argument):
    if len(args)>iarg:
        args= args[:iarg] + (argument, ) + args[iarg+1:]
    else:
        kwargs[name] = argument
    return args, kwargs


def inject_status(recursion_dict):
    def decorate(fun):
        @wraps(fun)
        def call(*args, **kwargs):
            chain_iters = intercept_argument(args, kwargs, 0, 'chain_iters')
            replacements = intercept_argument(args, kwargs, 1, 'replacements')
            
            run_string = make_run_string(replacements)
            this_converged = recursion_dict[run_string].get('This Converged', False)
            if not this_converged:
                replacers = recursion_dict[run_string].get('Replacers', {})
                if replacers:
                    chain_iters_new = {}
                    for (key,var) in chain_iters.items():
                        var = var[var.index(replacers[key]):]
                        chain_iters_new[key] = var
                    replacements = merge_dicts(replacements, replacers)
                args, kwargs = reinsert_argument(args, kwargs, 0, 'chain_iters', chain_iters_new)
                args, kwargs = reinsert_argument(args, kwargs, 1, 'replacements', replacements)
                return fun(*args, **kwargs)
            else:
                return {}
        return call
    return decorate

def static_vars(**kwargs):
    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func
    return decorate

def auto_conv_rerun(fun):
    """ Decorator for rerunning :func:`auto_conv`, until convergence is achieved in 
the first two calculations for each parameter. This is useful for cases where 
parameters are strongly correlated"""
    @wraps(fun)
    @static_vars(status={})
    def call(*args, **kwargs):
        converged_parameters = fun(*args, **kwargs)
        everything_converged=True
        for run_string in converged_parameters:
            replacements={}
            prevvarvalues={}
            call.status[run_string]=call.status.get(run_string, {})
            this_converged = call.status[run_string].get('This Converged', False)
            if not this_converged:
                this_converged = True
                for var_name in converged_parameters[run_string]:
                    replacements[var_name]=converged_parameters[run_string][var_name]['VarValue']
                    if converged_parameters[run_string][var_name]['iteration']>2:
                        this_converged = False
                        everything_converged = False
                    prevvarvalues[var_name]=converged_parameters[run_string][var_name]['PrevVarValue']
                if this_converged:
                    call.status[run_string]['Final replacements'] = replacements
                call.status[run_string]['This Converged'] = this_converged
                call.status[run_string]['Replacers'] = prevvarvalues
                call.status[run_string]['Run no'] = call.status[run_string].get('Run no', 0)+1
        import pprint
        pprint.pprint(call.status)
        if not everything_converged:
            auto_conv_handle = inject_status(call.status)(auto_conv_sub)
            kwargs.update({'auto_conv_sub': auto_conv_handle})
            converged_parameters=call(*args, **kwargs)
        else:
            converged_parameters=call.status
        return converged_parameters
    return call



#!/usr/bin/python
import pandas as pd

from runana.run import is_it_tuple
from runana.read_numbers import ignored
from runana import analyse


class SeqsDataFrame(pd.DataFrame):
    numparam = 'NumParam'
    numparamval = 'NumParamValue'

    @property
    def _constructor(self):
        return SeqsDataFrame

    def iterator(self):
        for numparam, data in self.iterator_outer():
            for column in data:
                dat = data[column]
                yield (numparam, column), dat

    def iterator_outer(self):
        for numparam in self.index.levels[0]:
            data = self.loc[(numparam)]
            data.sort_index(inplace=True)
            yield numparam, data

    def iterator_drop(self):
        for (numparam, column), dat in self.iterator():
            dat = dat.dropna()
            if not dat.empty:
                yield (numparam, column), dat

    def iterator_all(self):
        for (numparam, column), dat in self.iterator():
            for (numparamvalue, elem) in dat.iteritems():
                yield ((numparam, numparamvalue), column), elem

    def iterator_all_drop(self):
        for (numparam, column), dat in self.iterator_drop():
            for (numparamvalue, elem) in dat.iteritems():
                yield ((numparam, numparamvalue), column), elem

    def import_from_seq(self, seqsnew, varvals, inplace=False):
        """Converts the seqs object into a SeqsDataFrame"""
        seqsdf = self if inplace else self.copy()
        multiindx = pd.MultiIndex(levels=[[], []], codes=[[], []],
                                  names=[seqsdf.numparam, seqsdf.numparamval])
        seqsdf.set_index(multiindx, inplace=True)
        whatever_scalar = 'lol'
        list_ = list(iterate_seqs(seqsnew, varvals))
        for multi_idx, dir_ in list_:
            seqsdf.loc[multi_idx] = whatever_scalar
        # for multi_idx, dir_ in list_:
# It might seem strange to repeat the same command twice, and indeed it is.
# It seems that pandas unpacks a tuple of length 1 first time its inserted,
# but not the second time...
            seqsdf.loc[multi_idx] = dir_
            seqsdf.loc[multi_idx] = dir_
        if not inplace:
            return seqsdf

    def calc_reldiff(self):
        """ Calculate relative difference of values and numerical parameter values

        `(O2-O1)/(x2-x1)` where `O` are values and `x` are numerical parameters

        All numerical parameter values have to be scalar and numeric

        Returns a new SeqsDataFrame
        """
        import numpy as np
        data_out = self.copy()
        columns = list(self.columns)
        data_out = data_out.drop(columns=columns)
        for (numparam, column), data in self.iterator():
            data = data.reset_index(level=self.numparamval)
            relDiff = data.diff()
            RelErrorEstimate = relDiff[column]/relDiff[self.numparamval]
            RelErrorEstimate = RelErrorEstimate.apply(np.abs)
            RelErrorEstimate = pd.Series(RelErrorEstimate.values,
                                         data[self.numparamval])
            for numparamval in data[self.numparamval]:
                data_out.loc[(numparam, numparamval), str(column)+'_reldiff'] = RelErrorEstimate[numparamval]
        return data_out

    def calc_convergence(self):
        """ Calculate `(O2-O1)/O2*x2/(x2-x1)` where `O` are values and `x` are
 numerical parameters

        All numerical parameter values have to be scalar and numeric

        Returns a new SeqsDataFrame
        """
        import numpy as np
        try:
            data_out = self.copy()
            columns = list(self.columns)
            data_out = data_out.drop(columns=columns)
            for (numparam, column), data in self.iterator():
                data = data.reset_index(level=self.numparamval)
                relDiff = data.diff()/data
                RelErrorEstimate = relDiff[column]/relDiff[self.numparamval]
                RelErrorEstimate = RelErrorEstimate.apply(np.abs)
                RelErrorEstimate = pd.Series(RelErrorEstimate.values,
                                             data[self.numparamval])
                for numparamval in data[self.numparamval]:
                    data_out.loc[(numparam, numparamval), str(column)+'_conv'] = RelErrorEstimate[numparamval]
            return data_out
        except TypeError as e:
            print(str(e))
            raise TypeError("Make sure that "+self.numparamval+" and values in"
                            + " the SeqsDataFrame are all numerical and scalar")

    def calc_convergence_func(self, func):
        """ Calculate `func(O1,O2)*x2/(x2-x1)` where `O` are values and `x` are
 numerical parameters

        All numerical parameter values have to be scalar and numeric

        Returns a new SeqsDataFrame
        """
        import numpy as np
        data_out = self.copy()
        columns = list(self.columns)
        data_out = data_out.drop(columns=columns)
        for (numparam, column), data in self.iterator():
            # print('data data[1:]')
            # print(type(data))
            # print(data)
            # print(data.iloc[1:])
            # print(data.iloc[:-1])
            for (x1, O1), (x2, O2) in zip(data.iloc[:-1].iteritems(),
                                          data.iloc[1:].iteritems()):
                # idx = 
                # data_out.loc[idx] = np.abs(func(O1, O2)*x2/(x2-x1))
                try:
                    dat = np.abs(func(O1, O2)*x2/(x2-x1))
                    # print('dat', dat)
                except TypeError:
                    dat = np.nan
                data_out.loc[((numparam, x1), str(column)+'_conv')] = dat
        return data_out

    def plot_(self, outfile, logx=False, logy=False, grid=False,
              param_panda=None):
        """ Requires :mod:`numpy` and :mod:`matplotlib`"""
        from runana import matplotlib_managers as mplm
        import numpy as np
        with mplm.plot_manager(outfile=outfile) as pp:
            for numparam, data in self.iterator_outer():
                with mplm.single_ax_manager(pp=pp) as ax:
                    data.plot(ax=ax, alpha=0.8, marker='o')
                    ax.set_xlabel(numparam)
                    ax.legend(loc='best')
                    if grid:
                        ax.grid()
                    if logx:
                        ax.set_xscale('log')
                    if logy:
                        ax.set_yscale('log')
                        # ymin,ymax = ax.get_ylim()
                        ymin = np.nanmin(data.values)
                        ymax = np.nanmax(data.values)
                        ymin = np.power(10, np.floor(np.log10(ymin)))
                        ymax = np.power(10, np.ceil(np.log10(ymax)))
                        if np.isfinite(ymin) and np.isfinite(ymax):
                            ax.set_ylim([ymin, ymax])
                    if param_panda is not None:
                        param_series = param_panda.loc[(numparam)]
                        string = ' '.join(extract_interesting_vars(param_series, numparam))
                        ax.text(-0.1, 1.05, string, transform=ax.transAxes)


def iterate_seqs(seqsnew, varvals):
    for nameval, seq_lists in seqsnew.items():
        for idx, seq_list in enumerate(seq_lists):
            vals = dict((dir_, try_to_float(varvals[nameval][dir_][0]))
                        for dir_ in seq_list)
            for dir_, val in sorted(vals.items(),
                                    key=lambda x: try_to_float(x[1])):
                numparam = is_it_tuple(nameval[0])
                yield ((numparam, val), idx), dir_


# def import_from_double_var(double_var, varvals):
#     """ """
#     whatever_scalar = 0.1
#     double_var_pandas = {}
#     for namevals, seq_lists in double_var.items():
#         for idx, seq_list in enumerate(seq_lists):
#             df = pd.DataFrame()
#             df.index.name = is_it_tuple(namevals[0])
#             df.columns.name = is_it_tuple(namevals[1])
#             vals = dict((dir_, (varvals[namevals][dir_])) for dir_ in seq_list)
#             for dir_, val in sorted(vals.items(), key=lambda x: x[1]):
#                 df.loc[val[0], val[1]] = whatever_scalar
#                 df.loc[val[0], val[1]] = dir_
#             double_var_pandas.setdefault(namevals, []).append(df)
#     return double_var_pandas
def import_from_double_var(double_var, varvals):
    """ """
    whatever_scalar = 0.1
    double_var_pandas = {}
    for namevals, seq_list in double_var_iter(double_var):
        df = pd.DataFrame()
        df.index.name = is_it_tuple(namevals[0])
        df.columns.name = is_it_tuple(namevals[1])
        for val0, val1, dir_ in namevals_iter(namevals, seq_list, varvals):
            df.loc[val0, val1] = whatever_scalar
            df.loc[val0, val1] = dir_
        double_var_pandas.setdefault(namevals, []).append(df)
    return double_var_pandas


def double_var_vectors(double_var, varvals):
    double_var_out = {}
    for namevals, seq_list in double_var_iter(double_var):
        val0, val1, dirs = zip(*namevals_iter(namevals, seq_list, varvals))
        print(val0, val1, dirs)
        double_var_out.setdefault(namevals, []).append((val0, val1, dirs))
    return double_var_out


# def double_var_vectors(double_var, varvals):
#     """ """
#     double_var = {}
#     for namevals, seq_lists in double_var.items():
#         for seq_list in seq_lists:
#             var1 = []
#             var2 = []
#             dirs = []
#             vals = dict((dir_, (varvals[namevals][dir_])) for dir_ in seq_list)
#             for dir_, val in sorted(vals.items(), key=lambda x: x[1]):
#                 var1.append(val[0])
#                 var2.append(val[0])
#                 dirs.append(dir_)
#             double_var.setdefault(namevals, []).append((var1, var2, dirs))
#     return double_var


def double_var_iter(double_var):
    for namevals, seq_lists in double_var.items():
        for seq_list in seq_lists:
            yield namevals, seq_list


def namevals_iter(namevals, seq_list, varvals):
    vals = dict((dir_, (varvals[namevals][dir_])) for dir_ in seq_list)
    for dir_, val in sorted(vals.items(), key=lambda x: try_to_float(x[1])):
        yield val[0], val[1], dir_


def extract_interesting_vars(param_series, numparam):
    for column in param_series:
        paramdicts = param_series[column].dropna()
        if not paramdicts.empty:
            paramdict = paramdicts.iloc[0]
            for param_str in write_paramdict(paramdict, numparam):
                yield ''.join((str(column), ': ', param_str))


def write_paramdict(paramdict, ignore=None, connector='='):
    for field in paramdict:
        # if field != ignore:
        if ignore not in field:
            yield ''.join((str(is_it_tuple(field)), connector,
                           str(paramdict[field])))


def return_dict_element(dict_, error=KeyError):
    """ Returns a function that returns `dict_[arg]`, while ignoring `error`"""
    @ignored(error)
    def return_element(el):
        return dict_[el]
    return return_element


def try_to_float(str_):
    try:
        return float(str_)
    except (ValueError, TypeError):
        return str(str_)


def make_a_seq_panda(dict_w_params):
    """ Convenience function for finding sequences of data, and putting them
 in a Pandas structure """
    seqs, varvals = analyse.groupby_n_var_params(dict_w_params, 1)
    panda_data = SeqsDataFrame().import_from_seq(seqs, varvals)
    return panda_data


def drop_boring_columns(df):
    """ Drops all columns in `df` that has only 1 or less unique values.
Returns a new DataFrame """
    nunique = df.apply(pd.Series.nunique)
    cols_to_drop = nunique[nunique <= 1].index
    return df.drop(cols_to_drop, axis=1)



from collections import deque
from math import isnan
from functools import partial
from os import path

from runana.read_numbers import ignored
from runana.run import cwd, get_subdirs
from runana.input_file_handling import read_input_files_f90nml, string_or_iterable, superset


def collecting_loop_recursive(dir_, read_func,
                              look_for_these=["hostname.txt"],
                              ignore_these=["ignore"]):
    """Collect information from runs.

    Recursively locate directories under `dir_` containing any of the
    files in `look_for_these`, but leaves out directories that contain
    any files in `ignore_these`. In each of the located directories the
    function `read_func` is run with no arguments (the working directory
    is temporarily changed to the located directory).

    """
    subdirs = get_subdirs(dir_)
    for subdir in subdirs or []:
        asubdir = path.join(dir_, subdir)
        if any(path.exists(path.join(asubdir, fname)) for fname in look_for_these):
            if not any(path.exists(path.join(asubdir, fname)) for fname in ignore_these):
                with cwd(asubdir):
                    value = read_func()
                yield (subdir, ), value
        else:
            for dirs, vals in collecting_loop_recursive(path.join(dir_, subdir),
                                                        read_func):
                yield (subdir, )+dirs, vals


def read_input_files(workdir, indices=[], read_func=read_input_files_f90nml):
    """ Read information from input files.

    Recursively searches through all subdirectories of `workdir`.
    `read_func` is run in any directory containing a file named 'hostname.txt',
    and the result is stored in a :class:`ParamDict`, with the path in tuple-form
    as key. This :class:`ParamDict` is returned.

    If `indices` is given as a non-empty list, the indices in this argument
    will be used instead of recursive search

    Subdirectories of a directory with a 'hostname.txt' file are not searched.
    """
    paramdict = ParamDict()
    if indices:
        paramdict.from_list_of_indices(workdir, indices,
                                       read_func=read_func)
    else:
        paramdict.read(workdir, read_func=read_func)
    return paramdict


class ParamDict(dict):
    """ Dictionary that holds dictionaries of parameters """
    def read(self, workdir, read_func=read_input_files_f90nml):
        for index, result in collecting_loop_recursive(workdir, read_func):
            self[index] = result

    def from_list_of_indices(self, workdir, indices,
                             read_func=read_input_files_f90nml):
        for index in indices:
            with cwd(path.join(workdir, *index)):
                self[index] = read_func()

    def diff(self):
        """ Call :func:`dictdiff` on `ParamDict` object """
        dictdiff(self)

    def unpack_list_values(self):
        """ Takes any numerical parameter value in `ParamDict` object that is
        a list and packs it into individual slots with name numparam_name + idx

        Works in-place
        """
        for param_dict in self.values():
            spread_out_lists(param_dict)


def spread_out_lists(dict_):
    newdict = {}
    delete_keys = []
    for key, val in dict_.items():
        if isinstance(val, list):
            delete_keys += [key]
            for idx, elem in enumerate(val):
                if isinstance(key, tuple):
                    newkey = (key[0], str(key[1])+'_'+str(idx+1))
                else:
                    newkey = str(key)+'_'+str(idx+1)
                newdict[newkey] = elem
    dict_.update(newdict)
    for key in delete_keys:
        del dict_[key]


def make_collector_function(workdir, read_func, *args, **kwargs):
    """ Returns a function that runs `read_func(*args,**kwargs)` in the
directory that is the join of `workdir` and the argument to the function
    """
    return read_from_dir(roll_in_args(read_func, *args, **kwargs), workdir)


# @ignored(TypeError)
def collect(dir_, read_func):
    """ Switches to `dir_` and runs `read_func`"""
    with cwd(dir_):
        return read_func()


def roll_in_args(read_func, *args, **kwargs):
    read_func_no_args = partial(read_func, *args, **kwargs)
    read_func_subdir = partial(collect, read_func=read_func_no_args)
    return read_func_subdir


def compose2(f__, g__):
    def fg_(*a, **kw):
        return f__(g__(*a, **kw))
    return fg_


@ignored(TypeError)
def join_dirs(subdirs, workdir):
    dir_ = path.join(workdir,
                     *tuple(string_or_iterable(subdirs)))
    return dir_


def prepend_dir(workdir):
    """ Returns a function that takes a tuple of directories and returns the
combination of those into a path, with `workdir` prepended """
    return partial(join_dirs, workdir=workdir)


def read_from_dir(read_func, workdir):
    """ Composes `read_func` with :func:`prepend_dir(workdir)<prepend_dir>` """
    return compose2(read_func, prepend_dir(workdir))


class Seqs(dict):
    """ Seqeunces of related runs

    :param dict param_dicts: Dictionary containing dictionaries of parameters,
        in the form returned from e.g. :func:`collect_from_all`
    """
    def __init__(self, param_dicts, *args, **kwargs):
        super(Seqs, self).__init__(*args, **kwargs)
        keys = deque(param_dicts.keys())
        for index in list(keys):
            del keys[0]
            for key, indices in get_indices_dict(index,
                                                 param_dicts, keys).items():
                seqs_list = self.get(key, [])
                if all(not indices_sub(param_dicts, indices.values(),
                                       indices_seqs.values())
                       for indices_seqs in seqs_list):
                    self[key] = seqs_list + [indices]

    def iterator(self):
        for key in self:
            for indx, seq_list in enumerate(self[key]):
                yield key, indx, seq_list


class ChangedParams(dict):
    """ Parameters that changed. Derived from dict

    :param dict param_dicts: Dictionary containing dictionaries of parameters,
        in the form returned from e.g. :func:`collect_from_all`
    """
    def __init__(self, param_dicts, *args, **kwargs):
        super(ChangedParams, self).__init__(*args, **kwargs)
        ignore = []
        keys = list(param_dicts.keys())
        for idx in list(keys):
            del keys[0]
            for idx_compare in keys:
                diffs = find_diff_elements(param_dicts[idx],
                                           param_dicts[idx_compare])
                if len(diffs) == 0:
                    ignore.append(idx_compare)
                elif idx_compare not in ignore and idx not in ignore:
                    self[(idx, idx_compare)] = diffs

    def groupby_varname(self):
        """ Groups elements according to the name of the variable.

        Returns a dictionary containing the values of the variables, and
        another with all the pairs of runs"""
        varvalues = {}
        pairs = {}
        for dirs, varnameval in self.items():
            # varnames = frozenset(varnameval.keys())
            varnames = tuple(varnameval.keys())
            varvals = varnameval.values()
            value = dict(zip(dirs, zip(*varvals)))
            sdirs = set(dirs)
            pairs.setdefault(varnames, []).append(sdirs)
            varvalues.setdefault(varnames, {}).update(value)
        return varvalues, pairs


def groupby_params(dict_w_params, params):
    """ Groups `dict_w_params` according the given set of parameters. This
 can be used to e.g. seperate physical from numerical parameters. """
    grouped_dict = {}
    for key, item in dict_w_params.items():
        item_copy = item.copy()
        index_dict = {}
        for param in params:
            if param in item_copy:
                index_dict[param] = item_copy[param]
                del item_copy[param]
        index_tuple = tuple((key, index_dict[key])
                            for key in sorted(index_dict))
        grouped_dict.setdefault(index_tuple, {}).update({key: item_copy})
    return grouped_dict


def remove_nested_dict_keys(dict_w_params, remove_keys=[]):
    for dict_ in dict_w_params.values():
        for key in remove_keys:
            if key in dict_:
                del dict_[key]


def select_by_key_len(connected, length=1):
    return dict((key, list_) for key, list_ in connected.items()
                if len(key) == length)


def find_connected_components(pairs):
    connected = {}
    for key, value in pairs.items():
        connected[key] = conso(value)
    return connected


type_ignore_isnan = ignored(TypeError)(isnan)


def find_diff_elements(dict1, dict2):
    diffs = {}
    for key, value1 in dict1.items():
        if key in dict2:
            value2 = dict2[key]
            if value1 != value2 and not type_ignore_isnan(value1) and not type_ignore_isnan(value2):
                diffs[key] = (value1, value2)
        else:
            diffs[key] = (value1, None)
    return diffs


def catch_list_values(value):
    try:
        hash(value)
    except TypeError:
        value = tuple(value)
    return value


def get_indices_dict(idx, param_dicts, keys):
    indices_dict = {}
    for idx_compare in keys:
        for key, val, val_compare in get_index_for_all_but_one_changed(param_dicts[idx], param_dicts[idx_compare]):
            # if key not in indices_dict:
            #     indices_dict[key] = catch_list_values(val, idx)
            # indices_dict[key].update(catch_list_values(val_compare, idx_compare))
            indices_dict.setdefault(key, {catch_list_values(val): idx}).update({catch_list_values(val_compare): idx_compare})
    return indices_dict


def get_index_for_all_but_one_changed(nl1, nl2):
    for key, value in nl1.items():
        with ignored(KeyError):
            value2 = nl2[key]
            if value != value2:
                if dict_eq_ignore(nl1, nl2, [key]):
                    yield key, value, value2


def dict_eq_ignore(dict1, dict2, ignore_keys=()):
    dict1_flat = copy_w_ignore_keys(dict1, ignore_keys)
    dict2_flat = copy_w_ignore_keys(dict2, ignore_keys)
    return dict1_flat == dict2_flat


def copy_w_ignore_keys(dict_, ignore_keys=()):
    return dict((key, value) for key, value in dict_.items()
                if key not in ignore_keys)


def indices_sub(dicts, indices1, indices2):
    dict_list1 = list(dicts[indx] for indx in indices1)
    dict_list2 = list(dicts[indx] for indx in indices2)
    return issubset(dict_list2, dict_list1)


def issubset(list1, list2):
    """ Checks if list1 is subset of list2 """
    list3 = []
    for elem in list2:
        if elem in list1:
            list3.append(elem)
    return list2 == list3


# This function was inspired by nmltab
def dictdiff(alldicts):
    """ In-place removes all key:value pairs that are shared across all dicts

    :param dict alldicts: a dictionary containing dictionaries"""
    superdict = superset(alldicts)
    for key in superdict:
        all_dicts_have_key = all((key in dict_) for dict_ in alldicts.values())
        if all_dicts_have_key:
            value_is_same = all((superdict[key] == dict_[key])
                                for dict_ in alldicts.values())
            if value_is_same:
                for dict_ in alldicts.values():
                    del dict_[key]


# From: https://rosettacode.org/wiki/Set_consolidation#Python
def consolidate(sets):
    setlist = [s for s in sets if s]
    for i, s1 in enumerate(setlist):
        if s1:
            for s2 in setlist[i+1:]:
                intersection = s1.intersection(s2)
                if intersection:
                    s2.update(s1)
                    s1.clear()
                    s1 = s2
    return [s for s in setlist if s]


# From: https://rosettacode.org/wiki/Set_consolidation#Python
def conso(s):
    if len(s) < 2:
        return s

    r, b = [s[0]], conso(s[1:])
    for x in b:
        if r[0].intersection(x):
            r[0].update(x)
        else:
            r.append(x)
    return r


def groupby_n_var_params(dict_w_params, n_var_params=1):
    """ Convenience function for finding sets of data for which
 `n_var_params` parameters are changing """
    dict_w_params = dict_w_params.copy()
    dictdiff(dict_w_params)
    changedsparams = ChangedParams(dict_w_params)
    varvals, pairs = changedsparams.groupby_varname()
    connected = find_connected_components(pairs)
    nseqs = select_by_key_len(connected, length=n_var_params)
    return nseqs, varvals



from contextlib import contextmanager
try:
    @contextmanager
    def whatever():
        yield

    @whatever()
    def whateever():
        pass
except TypeError:
    from contextlib2 import contextmanager


@contextmanager
def ignored(*exceptions):
    try:
        yield
    except exceptions:
        pass


# ignored = contextmanager(ignored)

# def ignore_error(error=IOError, return_=None):
#     def ignore(func):
#         @wraps(func)
#         def wrapper(*args, **kwargs):
#             try:
#                 return func(*args, **kwargs)
#             except error:
#                 return return_
#         return wrapper
#     return ignore


ignore_missing_file = ignored(IOError)


@ignore_missing_file
def read_last_number_from_file(fname, pattern=''):
    number = None
    with open(fname) as stdout_file:
        for number in numbers_in_file_iterator(stdout_file, pattern=pattern):
            pass
    return number


@ignore_missing_file
def read_smallest_number_from_file(fname, pattern=''):
    with open(fname) as stdout_file:
        smallest = 100000
        for number in numbers_in_file_iterator(stdout_file, pattern=pattern):
            if number < smallest:
                smallest = number
    return smallest


@ignore_missing_file
def read_number_from_file(fname, inumber, pattern=''):
    with open(fname) as stdout_file:
        for indx, number in enumerate(numbers_in_file_iterator(stdout_file, pattern=pattern)):
            if indx == inumber-1:
                return number
    number = 'Not enough numbers in file'
    return number


@ignore_missing_file
def read_column_from_file(fname, icolumn, pattern=''):
    with open(fname) as stdout_file:
        for line in lines_in_file_iterator(stdout_file, pattern=pattern):
            for indx, word in enumerate(line.split()):
                if indx == icolumn-1:
                    return word


@ignore_missing_file
def read_entire_file(fname):
    with open(fname) as file_:
        return file_.readlines()


def lines_in_file_iterator(file_handle, pattern=''):
    for line in file_handle:
        if pattern in line:
            yield line


def words_in_file_iterator(file_handle, pattern=''):
    for line in lines_in_file_iterator(file_handle, pattern=pattern):
        for word in line.split():
            yield word


def numbers_in_file_iterator(file_handle, pattern=''):
    for word in words_in_file_iterator(file_handle, pattern=pattern):
        try:
            number = float(word)
            yield number
        except ValueError:
            pass


def split(delimiters, string, maxsplit=0):
    import re
    regexPattern = '|'.join(map(re.escape, delimiters))
    return re.split(regexPattern, string, maxsplit)


@ignore_missing_file
def read_file_sev_blocks(fname):
    with open(fname, 'r') as fil:
        blocks = [[[float(element) for element in line.split()]
                   for line in block.split('\n') if len(line) > 0]
                  for block in split(['\n \n', '\n\n'], fil.read())
                  if len(block) > 0]
    return blocks


def numpy_file_read(fname):
    import numpy as np
    try:
        a = np.loadtxt(fname)
    except IOError:
        a = None
    return a


def read_file_sev_blocks_c(fname):
    with open(fname, 'r') as fil:
        blocks = [[[num_c(element) for element in line.split()]
                   for line in block.split('\n') if len(line) > 0]
                  for block in split(['\n \n', '\n\n'], fil.read())
                  if len(block) > 0]
    return blocks


def read_file_super_blocks(fname):
    with open(fname, 'r') as fil:
        blocks = [[[[num_c(element) for element in line.split()]
                    for line in block.split('\n') if len(line) > 0]
                   for block in split(['\n \n', '\n\n'], super_block) if len(block) > 0]
                  for super_block in split(['\n \n \n', '\n\n\n'], fil.read()) if len(super_block) > 0]
    return blocks


def read_file_sev_blocks_new(fname):
    with open(fname, 'r') as fil:
        lines = [[num_c(field) for field in line.split()] for line in fil
                 if not line.lstrip()[:1] == '#']
    lines = split_list(lines)
    return lines


def read_file_sev_blocks_float(fname):
    with open(fname, 'r') as fil:
        lines = [[float(field) for field in line.split()] for line in fil
                 if not line.lstrip()[:1] == '#']
    return split_list(lines)


def read_file_sev_blocks_float_no_comment(fname):
    with open(fname, 'r') as fil:
        lines = [[float(field) for field in line.split()] for line in fil]
    return split_list(lines)


def split_list(list_):
    len_lines = list(map(len, list_))
    zero_idx = [i for i, x in enumerate(len_lines) if x == 0]
    if len(zero_idx) == 0:
        return list_
    else:
        if len(list_)-1 != zero_idx[-1]:
            zero_idx = zero_idx + [len(list_)]
        rangez = [(i+1, j) for i, j in zip(zero_idx, zero_idx[1:])]
        if zero_idx[0] != 0:
            rangez = [(0, zero_idx[0])] + rangez
        splitted_list = split_list([list_[i:j] for i, j in rangez])
        return splitted_list


def num_c(s):
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return complex(s)


def num(s):
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return s


def read_file_one_block(fname):
    with open(fname, 'r') as fil:
        block = [[num(element) for element in line.split()]
                 for line in fil.read().split('\n') if len(line) > 0]
    return block


def read_file_one_block_c(fname):
    with open(fname, 'r') as fil:
        block = [[num_c(element) for element in line.split()]
                 for line in fil.read().split('\n') if len(line) > 0]
    return block


def read_file_one_block_numpy(fname):
    import numpy as np
    block = read_file_one_block(fname)
    return np.asarray(block)



from shutil import copy

try:
    basestring          # Python 2.x
except NameError:
    basestring = str    # Python 3.x


def filter_inp_file_f90nml(inp_file_in, inp_file_out, replace_with_these):
    """ Replaces elements in *inp_file_in* and places the result in *inp_file_out*

    replace_with_these is a dict with entries of the form
    {'Name of parameter':*value to replace with*}

    This version works on namelist files using the f90nml package
    """
    import f90nml
    patch = patch_from_tupled_dict(replace_with_these)
    if patch:
        f90nml.patch(inp_file_in, patch, inp_file_out)
    else:
        copy(inp_file_in, inp_file_out)


def patch_from_tupled_dict(tupled_dict):
    """ Creates a nested dictionary that f90nml can interpret as a patch."""
    patch = {}
    for replace_this, replace_val in tupled_dict.items():
        group, field = replace_this
        group, field = group.lower(), field.lower()
        gdict = patch.get(group, {})
        gdict.update({field: replace_val})
        patch[group] = gdict
    return patch


def filter_inp_file_upname(inp_file_in, inp_file_out, replace_with_these):
    """ Replaces elements in *inp_file_in* and places the result in *inp_file_out*

    replace_with_these is a dict with entries of the form
    {'Name of parameter':*value to replace with*}

    This version replaces entries that is one line below a string matching
    *Name of parameter*, in the same position as the string
    """
    with open(inp_file_in, 'r') as input_file_in:
        input_data = input_file_in.readlines()
    replace_pos = dict(find_upname_positions(input_data, replace_with_these))
    output_data = '\n'.join(replace_lines(input_data, replace_pos))
    with open(inp_file_out, 'w') as input_file_out:
        input_file_out.write(output_data)


def find_upname_positions(input_data, replace_w_these):
    """ Finds the positions at which to do replacements. """
    for iline, line in enumerate(input_data):
        for replace_this, with_this in replace_w_these.items():
            for iword, word in enumerate(line.split()):
                if replace_this == word:
                    yield (iline+1, iword), with_this


def filter_inp_file_positional(inp_file_in, inp_file_out, replace_with_these):
    """ Replaces elements in *inp_file_in* and places the result in *inp_file_out*

    replace_with_these is a dict with entries of the form
    {(line number,field number):*value to replace with*}
    """
    with open(inp_file_in, 'r') as input_file_in:
        input_data = input_file_in.readlines()
    output_data = '\n'.join(replace_lines(input_data, replace_with_these))
    with open(inp_file_out, 'w') as input_file_out:
        input_file_out.write(output_data)


def replace_lines(input_data, replace_w_these):
    for iline, line in enumerate(input_data):
        lin = line.split()
        for (iline_rep, ifield_rep), replace_w_this in replace_w_these.items():
            if iline == iline_rep:
                lin[ifield_rep] = str(replace_w_this)
        yield '  '.join(lin)


#: Available input file filter functions
INP_FILE_FILTERS = {'f90nml': filter_inp_file_f90nml,
                    'upname': filter_inp_file_upname,
                    'positional': filter_inp_file_positional}


def read_positional_file(filename):
    """ Reads input file

    A dictionary is returned, the keys of which are the tuples `(iline,iword)`
    with the line and word index of the value. Each variable/name should be
    seperated by at least two whitespaces, a comma or a tab.
    """
    with open(filename, 'r') as input_file_in:
        input_data = input_file_in.readlines()
    return dict(index_line_words(input_data))


def index_line_words(input_list):
    for iline, line in enumerate(input_list):
        line = line.replace(',', '  ').replace('\t', '  ').split('  ')
        for iword, word in enumerate(line):
            yield (iline, iword), word


def read_upname_file(filename):
    """ Reads file in upname format

    In this format the names of variables are given on a line, with the values
    on the following line. Each variable/name should be seperated by at least
    two whitespaces or a tab. Each set of names/values lines should be
    seperated by at least one blank line
    """
    with open(filename, 'r') as input_file_in:
        input_data = input_file_in.readlines()
    previous_line = []
    data = {}
    for line in input_data:
        # Break line on tab or more than two consequtive whitespaces
        line = line.rstrip().replace(',', '  ').replace('\t', '  ').split('  ')
        line = [num_str(elem.strip()) for elem in line if elem]
        data.update(dict(zip(previous_line, line)))
        previous_line = line
    return data


def read_and_flatten_namelist(filename):
    import f90nml
    namelist = f90nml.read(filename)
    dict_ = dict(flat_iterator(namelist))
    return dict_


def read_input_files_f90nml(patterns=['*.nml'],
                            read_one_file=read_and_flatten_namelist):
    """ Read the all files matching `patterns` with :func:`f90nml.read`

    The namelists are flattened and supersetted, the resulting dict is
    returned
    """
    from itertools import chain
    import glob
    filenames = list(chain(*(glob.glob(pattern)
                             for pattern in string_or_iterable(patterns))))
    return read_list_of_input_files(filenames,
                                    read_one_file=read_one_file)
    # dicts = dict((filename, read_one_file(filename))
    #              for filename in filenames)
    # return superset_collisions(dicts)


def read_list_of_input_files(filenames,
                             read_one_file=read_and_flatten_namelist):
    """ Read the all files in list `filenames` with :func:`f90nml.read`

    The namelists are flattened and supersetted, the resulting dict is
    returned
    """
    dicts = dict((filename, read_one_file(filename))
                 for filename in filenames)
    return superset_collisions(dicts)


def read_input_files_upname(patterns=['*.inp']):
    """ Read the all files matching `patterns` with :func:`read_upname_file`

    The data from all the files is supersetted and the resulting dict is
    returned
    """
    return read_input_files_f90nml(patterns=patterns,
                                   read_one_file=read_upname_file)


def read_input_files_positional(patterns=['*.inp']):
    """ Read the all files matching `patterns` with :func:`read_positional_file`

    The data from all the files is supersetted and the resulting dict is
    returned
    """
    return read_input_files_f90nml(patterns=patterns,
                                   read_one_file=read_positional_file)


def flat_iterator(nml):
    """Iterator that returns the adress of an element as a 2-tuple,
    along with the element """
    for key, value in nml.items():
        for inner_key, inner_value in value.items():
            yield (key, inner_key), inner_value


def string_or_iterable(args):
    """ Generator that assumes args is either a string or an iterable. """
    if isinstance(args, basestring):
        yield args
    else:
        for arg in args:
            yield arg


# This function was inspired by nmltab
def superset(alldicts):
    """ Returns dict containing all keys from the dicts contained in `alldicts`

    :param dict alldicts: a dictionary containing dictionaries"""
    superdict = {}
    for dict_ in alldicts.values():
        superdict.update(dict_)
    return superdict


def superset_collisions(alldicts):
    """ Returns dict containing all keys from the dicts contained in `alldicts`

    :param dict alldicts: a dictionary containing dictionaries

    If two dictionaries contains the same key, the keys will be appended with
    the difference of the file names. """
    superdict = {}
    filenames = list(alldicts.keys())
    filediff_dict = dict(zip(filenames,
                             extract_noncommon_substrings(filenames)))
    for key, dict_ in alldicts.items():
        for innerkey, value in dict_.items():
            if innerkey in superdict and value != superdict[innerkey]:
                try:
                    innerkey += filediff_dict[key]
                except TypeError:
                    try:
                        # innerkey += (filediff_dict[key],)
                        innerkey = (innerkey[0],
                                    str(innerkey[1]) + filediff_dict[key])
                    except TypeError:
                        innerkey = str(innerkey) + filediff_dict[key]
            superdict[innerkey] = value
    return superdict


def num_str(s):
    """ Tries to convert input to integer, then tries float, then complex.
If all these fails the string is returned unchanged"""
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            try:
                return complex(s)
            except ValueError:
                return s


def extract_common_substrings(strings):
    """ Returns all substrings that are common between the strings in the list
`strings` as a single string.
    """
    common = strings[0]
    for string in strings:
        common = ''.join(ndiff_select(string, common))
    return common


def extract_noncommon_substrings(strings, common=None):
    """ Returns a list with parts of strings that are different.

    `common` is a string of all the common parts of `strings`, such as e.g.
    returned from func:`extract_common_substrings`. If set set to `None`,
    func:`extract_common_substrings` will be called to set this variable """
    if common is None:
        common = extract_common_substrings(strings)
    noncommon = [''.join(ndiff_select(string, common, '-'))
                 for string in strings]
    return noncommon


def ndiff_select(string, common, selector=' '):
    from difflib import ndiff
    for s in ndiff(string, common):
        if s[0] == selector:
            yield '{}'.format(s[-1])



import os
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="runana",
    version="0.1.5",
    author="Jens Svensmark",
    author_email="jenssss@uec.ac.jp",
    # description = ("An demonstration of how to create, document, and publish "
    #                                "to the cheese shop a5 pypi.org."),
    # license = "BSD",
    keywords="run analyse",
    # url = "http://packages.python.org/an_example_pypi_project",
    packages=['runana'],
    long_description=read('README.rst'),
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Utilities",
    ],
)



from __future__ import print_function

from os import getcwd, rename
from os.path import splitext, abspath, split
from os.path import join as pjoin
from contextlib import contextmanager

from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.backends.backend_svg import FigureCanvasSVG
from matplotlib.figure import Figure

from runana.run import makedir


class IndivFiles(object):
    """ Keeps track of png or svg files in a directory
    """
    def __init__(self, outfile, *args, **kwargs):
        basedir, file_ext = splitext(outfile)
        basedir = abspath(pjoin(getcwd(), basedir))
        makedir(basedir)
        self.cwd = getcwd()
        self.outfile = outfile
        self.basedir = basedir
        self.file_ext = file_ext
        self.args = args
        self.kwargs = kwargs
        self.page = 1
        if self.file_ext == ".png":
            self.canvas_func = FigureCanvas
        elif self.file_ext == ".svg":
            self.canvas_func = FigureCanvasSVG
        else:
            raise ValueError("File extension {} is not known".format(self.file_ext))

    def savefig(self, figure, **kwargs):
        try:
            fname = pjoin(self.basedir, "{}{}".format(self.page,
                                                      self.file_ext))
            if self.page == 1:
                fname = pjoin(self.cwd, self.outfile)
            elif self.page == 2:
                rename(pjoin(self.cwd, self.outfile),
                       pjoin(self.basedir, "{}{}".format(1,
                                                          self.file_ext)))
            orig_canvas = figure.canvas
            figure.canvas = self.canvas_func(figure)
            figure.savefig(fname, format=self.file_ext[1:], **kwargs)
            self.page += 1
        finally:
            figure.canvas = orig_canvas

    def get_pagecount(self):
        return self.page

    def close(self):
        """ No clean up needed """
        pass


class webmMovie(IndivFiles):

    def __init__(self, outfile, *args, **kwargs):
        self.outfile_video = outfile
        self.bitrate = kwargs.pop("bitrate", "1M")
        self.reverse = kwargs.pop("reverse", False)
        __, file_ext = splitext(self.outfile_video)
        self.file_ext_video = file_ext
        for file_ext in [".webm", ".mp4"]:
            outfile = outfile.replace(file_ext, ".png")
        super(webmMovie, self).__init__(outfile, *args, **kwargs)

    def print_args(self, args, fname, mode="w"):
        with open(pjoin(self.basedir, fname), mode) as file_:
            file_.write(" ".join(args+["\n"]))

    def close(self):
        from subprocess import call
        outfile_video = self.outfile_video
        if self.file_ext_video == ".webm":
            options = ["-y", "-c:v", "libvpx-vp9", "-b:v", self.bitrate]
        elif self.file_ext_video == ".mp4":
            options = ["-y", "-c:v", "libx264", "-b:v", self.bitrate]
            options += ["-pix_fmt", "yuv420p"]
        else:
            raise ValueError("Unsupported file_ext: " + self.file_ext_video)
        args = ["ffmpeg"]
        args.extend(["-i", pjoin(self.basedir, "%d.png")])
        # lines = []
        # filenames = []
        nfiles = self.page-1
        for i in range(nfiles):
            filename = pjoin(self.basedir, "{}.png".format(i+1))
            if self.reverse:
                filename2 = pjoin(self.basedir, "{}.png".format(2*nfiles-i))
                from os import remove
                from shutil import copy
                from runana.run import ignored
                with ignored(OSError):
                    remove(filename2)
                copy(filename, filename2)
        # filenames = []
        # for i in range(self.page-1):
        #     filename = pjoin(self.basedir, "{}.png".format(i+1))
        #     filenames.append(filename)
        #     lines.append("file '{}'\n".format(filename))
        # if self.reverse:
        #     filenames += reversed(filenames)
        # input_files = pjoin(self.basedir, "input_files.txt")
        # with open(input_files, "w") as file_:
        #     file_.write("".join(lines))
        #     if self.reverse:
        #         file_.write("".join(reversed(lines)))
        # args.extend(["-f", "concat",  "-i", input_files])
        # nested_list = [["-i", filename] for filename in filenames]
        # args.extend([arg for filename in filenames
        #              for arg in ["-i", filename]])
        # args.extend([item for sub in nested_list for item in sub])
        args.extend(options)
        args.append(outfile_video)
        self.print_args(args, "ffmpeg_args.txt")
        call(args)
        # with open(pjoin(self.basedir, "ffmpeg_args.txt"), "w") as file_:
        #     file_.write(" ".join(args+["\n"]))
        # if self.reverse:
        #     reverse_fname = "reverse.webm"
        #     args = ["ffmpeg", "-i", outfile_video,
        #             "-vf", "reverse"]
        #     args.extend(options)
        #     args.append(reverse_fname)
        #     call(args)
        #     self.print_args(args, "ffmpeg_args.txt", mode="aw")
        #     video_list = pjoin(self.basedir, "video_list.txt")
        #     self.print_args(["file", outfile_video], video_list, mode="w")
        #     self.print_args(["file", reverse_fname], video_list, mode="aw")
        #     args = ["ffmpeg", "-f", "concat",
        #             "-i", video_list,
        #             reverse_fname]
        #     # args.extend(options)
        #     args.extend(options)
        #     args.append("reverse_combined.webm")
        #     call(args)
        #     self.print_args(args, "ffmpeg_args.txt", mode="aw")


class plot_manager(object):
    """ Creates pdf file using :func:`matplotlib.backends.backend_pdf.PdfPages`
 and returns the handle to this pdf.

    `*args` and `*kwargs` are passed to `PdfPages`
    """
    def __init__(self, outfile, print_outfile=True, *args, **kwargs):
        self.outfile = outfile
        self.print_outfile = print_outfile
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        _, file_ext = splitext(self.outfile)
        if file_ext == ".pdf":
            self.pp = PdfPages(self.outfile, *self.args, **self.kwargs)
        elif file_ext == ".svg" or file_ext == ".png":
            self.pp = IndivFiles(self.outfile, *self.args, **self.kwargs)
        elif file_ext == ".webm" or file_ext == ".mp4":
            self.pp = webmMovie(self.outfile, *self.args, **self.kwargs)
        else:
            raise ValueError("Unknown file extension {}".format(file_ext))
        return self.pp

    def __exit__(self, type, value, traceback):
        self.pp.close()
        if self.print_outfile:
            print(self.outfile)


class single_fig_manager(object):
    """ Create a :class:`matplotlib.figure.Figure`

    `args` and `kwargs` are passed to the :class:`Figure` constructor

    :param matplotlib.backends.backend_pdf.PdfPages pp: handle that has a
        savefig method
    """
    def __init__(self, pp, *args, **kwargs):
        self.pp = pp
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        self.fig = Figure(*self.args, **self.kwargs)
        FigureCanvas(self.fig)
        return self.fig

    def __exit__(self, type, value, traceback):
        self.pp.savefig(self.fig, transparent=True)
        # self.pp.savefig(self.fig, transparent=True, bbox_inches="tight")
        self.fig.clear


class single_ax_manager(single_fig_manager):
    """ Create an axis with a single subplot in a `Figure` object.

    Subclassed from :class:`single_fig_manager`, and all the arguments are the
    same
    """
    def __enter__(self):
        self.fig = super(single_ax_manager, self).__enter__()
        ax = self.fig.add_subplot(1, 1, 1)
        return ax


@contextmanager
def plot_ax_manager(outfile, *args, **kwargs):
    """ Create a pdf with name `outfile` containing one plot"""
    with plot_manager(outfile, *args, **kwargs) as pp:
        with single_ax_manager(pp) as ax:
            yield ax
